%% pmedf_putLinesInHeader --- it works like pmedf_putInHeader(), but multiple
%% keys and values can be inserted into the header, thus being faster and more
%% comfortable if multiple keys are in a structure or in a cell array.
%%
%% Usage:
%%	header = pmedf_putLinesInHeader( header, keys, values [, pos] )
%%	header = pmedf_putLinesInHeader( header, keys_struct [, pos] )
%%
%% Examples:
%%	header = pmedf_putLinesInHeader( header, {'energy', 'note'}, {12.3, 'today'}, 16 );
%%
%%	keys{1}='energy'; values{1}=12.3; keys{2}='note'; values{2}='today';
%%	header = pmedf_putLinesInHeader( header, keys, values );
%%
%%	hs.energy = 12.3;  hs.note = 'today';
%%	header = pmedf_putLinesInHeader( header, hs );
%%
%%	hs = pmedf_splitHeader(header); hs.new_energy = 24.6;
%%	header = pmedf_putLinesInHeader( '', hs, 16 );
%%
%% Author: Petr Mikulik
%% Version: 19. 1. 2017
%% History:
%%	    January 2017:
%%		First version.

function header = pmedf_putLinesInHeader( header, keys, values, pos )

if nargin==0 || nargin>4 || ~ischar(header)
    error('Syntax: pmedf_putLinesInHeader(...)');
end
if isempty(header)
    header = pmedf_emptyHeader();
end

if isstruct(keys)
    if nargin==4 error('too many input parameters for the keys structure');
    elseif nargin==3 pos=values;
    else pos=0;
    end
    newkeys = fieldnames(keys);
    nkeys = numel(newkeys);
else
    if nargin==3 pos=0; end
    if ~iscell(keys) || ~iscell(values)
	error('keys and values must be cell arrays, or keys must be a structure');
    end
    if numel(keys) ~= numel(values)
	error('cell arrays keys and values must have the same number of items');
    end
    nkeys = numel(keys);
end

for k=1:nkeys
    if isstruct(keys)
	key = newkeys{k};
	new_value = getfield(keys, newkeys{k});
    else
	key = keys{k};
	new_value = values{k};
    end
    % Search for "\nkey ":
    key2 = [sprintf('\n') key ' '];
    value = strfind(header, key2);
    % Search for "\nkey=":
    if isempty(value)
	key2 = sprintf(['\n' key '=']);
	value = strfind(header, key2);
	if isempty(value) % key not found, use the default "\nkey "
	    key2 = [sprintf('\n') key ' '];
	end
    end
    key = key2;

    if isnumeric(new_value) new_value = sprintf('%g', new_value); end

    orig_size = length(header);

    if ~isempty(value) % the key already exists
	header_beg = [header(1:value(1)-1)];
	header = header(value(1):length(header));
	p = strfind(header, sprintf(';\n'));
	line = header(1:p(1)-1);
	header = header(p(1)+2:length(header)-2); % don't copy the trailing '}\n'
	p = strfind(line, '=');
	line = line(1:p(1));
	header = [header_beg line ' ' new_value sprintf(' ;\n') header];
	% strip the trailing spaces after the last line
	p = strfind(header, sprintf(';\n'));
	header = header(1:p(end)+1);
    else % the key does not exist yet
	p = strfind(header, sprintf(';\n'));
	if isempty(p)
	    p = strfind(header, sprintf('}\n'));
	    if isempty(p) p = strfind(header, sprintf('\n')); end
	    if isempty(p) p = 1; end;
	end
	header = header(1:p(end));
	poss = pos - length(key) - 1; % how many spaces to add
	if poss > 1
	    % spaces = [repmat(' ', 1, poss+1)]; % slower
	    % header = [header key spaces '= ' new_value sprintf(' ;\n')];
	    header = [header key sprintf('%*s= %s;\n', poss+1, ' ', new_value)];
	else
	    header = [header key '= ' new_value sprintf(' ;\n')];
	end
    end
end

% now extend the header to its original size or extend it
l = length(header);
if l+2 > orig_size % extend the header
    orig_size = 512*ceil(l/512);
end
% header = [header repmat(' ', 1, orig_size-l-2) sprintf('}\n')]; % slower
header = [header sprintf('%*s}\n', orig_size-l-2, ' ')]; % faster

% eof pmedf_putInHeader.m
