%% pmedf_splitHeader --- split (parse) header string into a structure.
%%
%% Each header key ('key = ...') becomes a field in the output
%% structure ('hs.key').
%% Number values are converted to double (e.g. Dim_1=1024),
%% strings stay strings (e.g. DataType='UnsignedShort').
%% Key can optionally change the letter case.
%%
%% Syntax:
%%	h = pmedf_read('sample_0001.edf');
%%      hedf = pmedf_splitHeader(h [, opt]);
%% where optional opt 'keepmixed' (default), 'tolower' or 'toupper' can change
%% the case of keys in the output field ('hedf.DataType', 'hedf.datatype' or
%% 'hedf.DATATYPE').
%%
%% Examples:
%%      hedf = pmedf_splitHeader(h);
%%      hedf = pmedf_splitHeader(h, 'tolower');
%%
%% Author: Petr Mikulik
%% Version: 24. 2. 2017
%% History:
%%	February 2017:
%%	    Octave allows any character in structure field names, while Matlab
%%	    allows only [a-zA-Z0-0-_]. Thus allow only those of Matlab
%%	    (well, found rarely in edf headers).
%%	January 2017:
%%	    Octave strsplit() changed to strfind() for Matlab compatibility.

function hs = pmedf_splitHeader ( h, opt )

if (nargin<1 || nargin>3)
    fprintf('Usage:\n');
    fprintf('  hedf = pmedf_splitHeader(header [, opt])\n');
    return
end
if nargin==1 opt='M';
elseif strcmp(opt, 'keepmixed') opt='M';
elseif strcmp(opt, 'tolower') opt='L';
elseif strcmp(opt, 'toupper') opt='U';
else error('Wrong case option: expecting "keepmixed", "tolower", "toupper"');
end

% There are limited string functions in Matlab compared to Octave.
is_octave = (exist('OCTAVE_VERSION', 'builtin')==5);
want_strsplit = 0; % elegant code for Octave for the same as for Matlab?

hh = h; % remove leading {\n
if is_octave % Octave (fast find of the 1st string)
    k = index(h, sprintf('{\n'));
else % Matlab (finds all occurencies, even though we need just the first one)
    k = strfind(h, sprintf('{\n'));
    if isempty(k) k=0; else k=k(1); end
end
if (k==1) hh = h(3:end); end
if want_strsplit
    hh = strsplit(hh, sprintf(';\n')); % split the header
    len_hh = length(hh);
else
    eoln = sprintf(';\n');
    leneoln = length(eoln);
    ihh = strfind(hh, eoln); % string find within the header
    ihh = [1-leneoln ihh length(ihh)+leneoln];
    len_hh = length(ihh)-1;
end

hs = struct; % en empty structure

for k=1:len_hh
    if want_strsplit
	s = hh{k};
    else
	s = hh(ihh(k)+leneoln:ihh(k+1)-leneoln);
    end
    p = strfind(s, '=');
    if isempty(p) continue; end
    key = s(1:p(1)-1);
    value = s(p(1)+1:end);
    % trim trailing spaces
    key = strtrim(key);
    value = strtrim(value);
    x = str2double(value);
    if is_octave % Octave
	if opt=='L' key=tolower(key); elseif opt=='U' key=toupper(key); end
    else % Matlab
	if opt=='L' key=lower(key); elseif opt=='U' key=upper(key); end
    end
    key = regexprep(key, '[^\w-_]', ''); % Matlab allows only [a-zA-Z0-0-_] in structure field names
    if isnan(x)
	hs = setfield (hs, key, value);
    else
	hs = setfield (hs, key, x);
    end
end

% eof pmedf_splitHeader.m

