%% pmedf_splitMnePos --- split (parse) strings with counter and motor mnemonics
%% and positions into a structure. 
%%
%% Syntax:
%%     hout = pmedf_splitMnePos ( h1 [,h1 [, opt]] );
%%       h1 can be either edf header string or header structure.
%%
%%     hout = pmedf_splitMnePos ( h1 );
%%       Returns structure with fields of counter and motor names and their values.
%%
%%     h2 = pmedf_splitMnePos(h1, h2);
%%       Adds counter and motor names and their values into the (edf) structure h2.
%%
%%     h2 = pmedf_splitMnePos(h1, h2, opt);
%%       Adds counter, motor or both keys and values into the (edf) structure h2,
%%       for opt being 'counter', 'motor' or 'both'.
%%
%% Examples:
%%
%% h  = pmedf_read('sample_0001.edf');
%% hs = pmedf_splitHeader(h);
%%
%% z = pmedf_splitMnePos(h);         % get motor and counter names
%% z = pmedf_splitMnePos(hs);        % same as above
%% hs = pmedf_splitMnePos(h, hs);    % same as above
%% hs = pmedf_splitMnePos(hs, hs);   % same as above
%% cs = pmedf_splitMnePos(h,  [], 'counters');  % only counters
%% hs = pmedf_splitMnePos(h,  hs, 'counters');
%% ms = pmedf_splitMnePos(hs, [], 'motors');    % only motors
%% hs = pmedf_splitMnePos(hs, hs, 'motors');
%%
%% Author: Petr Mikulik
%% Version: 13. 1. 2017
%% History:
%%	January 2017:
%%	    Octave strsplit() changed to strfind() for Matlab compatibility.

function hout = pmedf_splitMnePos ( h1, h2, opt )

if (nargin<1 || nargin>4)
    error('Usage: pmedf_splitMnePos(h1, [, h2 [, mne]])');
end

if nargin==1
    % if ~isstruct(h1) h1 = pmedf_splitHeader(h1); end
    opt = 'B';
    hout = struct;
elseif nargin==2
    % if ~isstruct(h1) h1 = pmedf_splitHeader(h1); end
    if isstruct(h2) hout = h2; else hout = struct; end
    opt = 'B';
elseif nargin==3
    % if ~isstruct(h1) h1 = pmedf_splitHeader(h1); end
    if isstruct(h2) hout = h2; else hout = struct; end
    if strcmp(opt, 'counters') opt='C';
    elseif strcmp(opt, 'motors') opt='M';
    elseif strcmp(opt, 'both') opt='B';
    else error('Expecting option "counters", "motors" or "both"');
    end
end

if ~isstruct(h1)
    if ~ischar(h1)
	error('First parameter must be header structure or header string');
    end
    hh.counter_mne = pmedf_findInHeader(h1, 'counter_mne', 'string');
    hh.counter_pos = pmedf_findInHeader(h1, 'counter_pos', 'string');
    hh.motor_mne = pmedf_findInHeader(h1, 'motor_mne', 'string');
    hh.motor_pos = pmedf_findInHeader(h1, 'motor_pos', 'string');
    h1 = hh;
end

if (opt=='B' || opt=='C')
    imne = strfind(h1.counter_mne, ' ');
    ipos = strfind(h1.counter_pos, ' ');
    imne = [0 imne length(h1.counter_mne)+1];
    ipos = [0 ipos length(h1.counter_pos)+1];
    for k=1:length(imne)-1
	mne = h1.counter_mne(imne(k)+1:imne(k+1)-1);
	pos = h1.counter_pos(ipos(k)+1:ipos(k+1)-1);
	hout = setfield(hout, mne, str2double(pos));
    end
end

if (opt=='B' || opt=='M')
    imne = strfind(h1.motor_mne, ' ');
    ipos = strfind(h1.motor_pos, ' ');
    imne = [0 imne length(h1.motor_mne)+1];
    ipos = [0 ipos length(h1.motor_pos)+1];
    for k=1:length(imne)-1
	mne = h1.motor_mne(imne(k)+1:imne(k+1)-1);
	pos = h1.motor_pos(ipos(k)+1:ipos(k+1)-1);
	hout = setfield(hout, mne, str2double(pos));
    end
end

% Tech note: nice strplit() solution works in Octave but not in Matlab which
% lacks this useful function.
%if (opt=='B' || opt=='C')
%    mne = strsplit(h1.counter_mne, ' ');
%    pos = strsplit(h1.counter_pos, ' ');
%    for k=1:length(mne)
%	hout = setfield(hout, mne{k}, str2double(pos{k}));
%    end
%end
%
%if (opt=='B' || opt=='M')
%    mne = strsplit(h1.motor_mne, ' ');
%    pos = strsplit(h1.motor_pos, ' ');
%    for k=1:length(mne)
%	hout = setfield(hout, mne{k}, str2double(pos{k}));
%    end
%end

% eof pmedf_splitMnePos.m

